////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2010-2013 60East Technologies LLC, All Rights Reserved.
//
// This computer software is owned by 60East Technologies LLC and is
// protected by U.S. copyright laws and other laws and by international
// treaties.  This computer software is furnished by 60East Technologies
// LLC pursuant to a written license agreement and may be used, copied,
// transmitted, and stored only in accordance with the terms of such
// license agreement and with the inclusion of the above copyright notice.
// This computer software or any other copies thereof may not be provided
// or otherwise made available to any other person.
//
// U.S. Government Restricted Rights.  This computer software: (a) was
// developed at private expense and is in all respects the proprietary
// information of 60East Technologies LLC; (b) was not developed with
// government funds; (c) is a trade secret of 60East Technologies LLC
// for all purposes of the Freedom of Information Act; and (d) is a
// commercial item and thus, pursuant to Section 12.212 of the Federal
// Acquisition Regulations (FAR) and DFAR Supplement Section 227.7202,
// Government’s use, duplication or disclosure of the computer software
// is subject to the restrictions set forth by 60East Technologies LLC.
//
////////////////////////////////////////////////////////////////////////////

#include <amps.h>
#include <HAClient.hpp>
#include <DefaultServerChooser.hpp>
#include <vector>
#include <map>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <fstream>
#include <limits>
#include <stdlib.h>

//
// class bridge_server_chooser
//
// Overrides DefaultServerChooser to add logging when a connection
// becomes unavailable.
//
class bridge_server_chooser : public AMPS::DefaultServerChooser
{
    // number of connection failures before we log to stderr.
    static const unsigned NUM_FAILURES_LOGGING_THRESHOLD = 1;

    // Counts the number of failures since last success
    unsigned _failures;
public:
    bridge_server_chooser() : _failures(0) { }

    // reportFailure is called by the HAClient whenever a connection
    // is attempted and fails.  We log every NUM_FAILURES_LOGGING_THRESHOLD.
    void reportFailure(const AMPS::AMPSException& ex_,
                       const AMPS::ConnectionInfo &connectionInfo_)
    {
        ++_failures;
        if(_failures && _failures % NUM_FAILURES_LOGGING_THRESHOLD ==0)
        {
            std::cerr <<
                      "warning: unable to connect to "
                      << connectionInfo_.find("client.uri")->second << ".  Retrying.  Most recent error was: " <<
                      ex_.what() << std::endl;
        }
    }

    // reportSuccess is called by HAClient when a connection succeeds.
    void reportSuccess(const AMPS::ConnectionInfo &info_)
    {
        if(_failures >= NUM_FAILURES_LOGGING_THRESHOLD)
        {
            std::cerr << "connectivity to " << info_.find("client.uri")->second << " restored." << std::endl;
        }
        _failures = 0;
    }
};

//
// class bridge_subscription
//
// bridge_subscription is a single source+destination pair.
// We use member variables configured elsewhere to determine what
// server to connect to, and what topics to subscribe/publish.
//
class bridge_subscription
{
public:
    bridge_subscription(void)
    {
        ;
    }

    bridge_subscription(const std::string& sourceURI_,
                        const std::string& sourceTopic_,
                        const std::string& sourceFilter_,
                        const std::string& destinationURI_,
                        const std::string& destinationTopic_,
                        bool  bookmark_)
        : _sourceURI(sourceURI_),
          _sourceTopic(sourceTopic_),
          _sourceFilter(sourceFilter_),
          _destinationURI(destinationURI_),
          _destinationTopic(destinationTopic_),
          _bookmark(bookmark_)
    {
        ;
    }

    // we require at least a URI and Topic for both sides.
    bool good() const
    {
        return !_sourceURI.empty() && !_sourceTopic.empty() &&
               !_destinationURI.empty() && !_destinationTopic.empty();
    }
    std::string    _sourceURI;
    std::string    _sourceTopic;
    std::string    _sourceFilter;
    AMPS::HAClient _sourceClient;
    AMPS::BookmarkStore _sourceBookmarkStore;

    std::string    _destinationURI;
    std::string    _destinationTopic;
    AMPS::HAClient _destinationClient;

    bool           _bookmark;
    std::string _subId;

    AMPS::Message _message;
};

typedef std::vector<bridge_subscription> bridge_subscriptions;

void bridgeMessageHandler(const AMPS::Message& message_,void *userData_);
void bookmarkBridgeMessageHandler(const AMPS::Message& message_,void *userData_);

class amps_bridge
{
    static const unsigned COMMAND_TIMEOUT = 100000;
public:
    amps_bridge(void)
        : _messageCount(0)
    {
        ;
    }

    void run(const bridge_subscriptions& bridgeSubscriptions_)
    {
        _subscriptions = bridgeSubscriptions_;
        for (size_t i = 0; i < _subscriptions.size(); ++i)
        {
            bridge_subscription& subscription = _subscriptions[i];

            // Find the source and destination clients.
            subscription._destinationClient = _getClient(subscription._destinationURI, subscription._bookmark);
            subscription._sourceClient = _getClient(subscription._sourceURI, subscription._bookmark);

            if(subscription._bookmark)
            {
                // When using a transaction log, we issue a bookmark subscription.
                // Since we need to refer to the bookmark store in order to "discard" in the 
                // message handler, we store a reference to it, here.
                subscription._sourceBookmarkStore = subscription._sourceClient.getBookmarkStore();

                // Subscribe with the specified topic and filter.
                // The special value BOOKMARK_NOW instructs AMPS to begin our subscription from the current point-in-time.
                // make a unique subscription id, and remember it:
                // these are used for determining 'how' we saw a particular
                // message, since we might have multiple subscriptions for the same topic
                std::ostringstream subIdBuilder;
                subIdBuilder << "sub-" << i;
                subscription._subId = subIdBuilder.str();
                subscription._subId.resize(AMPS::Message::IdentifierLength, ' ');
                subscription._sourceClient.bookmarkSubscribe(
                    AMPS::MessageHandler(&bookmarkBridgeMessageHandler,&subscription),subscription._sourceTopic,COMMAND_TIMEOUT,
                    AMPS::Client::BOOKMARK_NOW(),subscription._sourceFilter,"", subscription._subId);
            }
            else
            {
                // In non-logged subscribe/publish, we gain a little extra performance by
                // recycling a message that's already set with the topic and command.
                subscription._message.setTopic(subscription._destinationTopic);
                subscription._message.setCommand("publish");
                subscription._sourceClient.subscribe(
                    AMPS::MessageHandler(&bridgeMessageHandler,&subscription),subscription._sourceTopic,
                    COMMAND_TIMEOUT,subscription._sourceFilter);
            }
        }
    }

    // TODO: Replace this filter with any additional processing you'd like done.
    // the default copies data from input to output as-is.
    inline void filter(const AMPS::Message& incoming_, std::string& outgoing_, const bridge_subscription* subscription_)
    {
        outgoing_=incoming_.getData();
    }

    void bridge(const AMPS::Message& message_,bridge_subscription* subscription_)
    {
        ++_messageCount;
        std::string output;
        try
        {
            filter(message_, output, subscription_);
        }
        catch (const std::exception &e)
        {
            std::cerr << "unhandled exception in filter function: " << e.what() << std::endl;
            exit(-1);
        }
        catch(...)
        {
            std::cerr << "unhandled exception of unknown type, in filter function." << std::endl;
            exit(-1);
        }

        subscription_->_message.setData(output);
        subscription_->_destinationClient.send(subscription_->_message);
    }

    void bookmark_bridge(const AMPS::Message& message_,bridge_subscription* subscription_)
    {
        ++_messageCount;
        std::string output;
        // if filter throws an exception, we will exit immediately, no discard or publish.
        try
        {
            filter(message_, output, subscription_);
        }
        catch (const std::exception &e)
        {
            std::cerr << "unhandled exception in filter function: " << e.what() << std::endl;
            exit(-1);
        }
        catch(...)
        {
            std::cerr << "unhandled exception of unknown type, in filter function." << std::endl;
            exit(-1);
        }

        try
        {
            // don't publish empty messages.
            if(!output.empty())
            {
                subscription_->_destinationClient.publish(subscription_->_destinationTopic,
                        output);
            }
        }
        catch (const AMPS::StoreException &ex)
        {
            // an error has occured writing to the local store.  terminate.
            std::cerr << "error writing to local store: " << ex.what() << std::endl;
            exit(-1);
        }
        catch (const AMPS::AMPSException &ex)
        {
            std::cerr << "error occured while publishing: " << ex.what() << std::endl;
            exit(-1);
        }

        // For HA subscriptions, we must explicitly "discard" each message once
        // we're finished processing it.
        subscription_->_sourceBookmarkStore.discard(subscription_->_subId.c_str(),
               message_.getBookmarkSeqNo());
    }

    size_t messageCount() const
    {
        return _messageCount;
    }

protected:

    // We maintain a map of client name to client,
    // separating HA and non-HA
    AMPS::HAClient _getClient(const std::string& uri_, bool isBookmark_)
    {
        std::string clientName = uri_ + (isBookmark_?"_bk":"");
        const client_map::iterator it = _clientMap.find(clientName);
        if (it == _clientMap.end())
        {
            AMPS::ServerChooser chooser(new bridge_server_chooser());
            chooser.add(uri_);

            AMPS::HAClient client(clientName);
            if(isBookmark_)
            {
                client = AMPS::HAClient::createMemoryBacked(clientName);
            }

            _clientMap.insert(client_map::value_type(clientName, client));
            client.setServerChooser(chooser);
            client.connectAndLogon();
            return client;
        }
        else
        {
            return it->second;
        }
    }

    typedef std::map<std::string,AMPS::HAClient> client_map;

    bridge_subscriptions _subscriptions;
    client_map           _clientMap;
    volatile size_t      _messageCount;
};

amps_bridge bridge;

void bookmarkBridgeMessageHandler(const AMPS::Message& message_, void *userData_)
{
    bridge.bookmark_bridge(message_,(bridge_subscription*)userData_);
}
void bridgeMessageHandler(const AMPS::Message& message_, void *userData_)
{
    bridge.bridge(message_,(bridge_subscription*)userData_);
}

// reads a single bridge subscription from the csv file
// format:
//   source-uri,topic,dest-uri,topic,[filter],[bookmark]
// blank, empty, and lines beginning with '#' are skipped.
std::istream& operator>>(std::istream& csvFile, bridge_subscription& subscription)
{
    std::string line;
    if(!getline(csvFile, line) || line.empty() || line[0] == '#' || line[0]==' ')
        return csvFile;
    std::string bookmark;
    std::string* ptrs[] = { &subscription._sourceURI, &subscription._sourceTopic,
                            &subscription._destinationURI, &subscription._destinationTopic, &subscription._sourceFilter, &bookmark, NULL
                          };
    size_t index = 0;
    for(unsigned i = 0; index!=std::string::npos && ptrs[i]!=NULL; i++)
    {
        size_t oldindex = index;
        index = line.find(',',index);
        *ptrs[i] = line.substr(oldindex, index-oldindex);
        if(index == std::string::npos) break;
        ++index;
    }
    // empty, n.*, f.* or 0.* mean "false", any other text means true
    subscription._bookmark = !bookmark.empty() && bookmark[0] !='n' && bookmark[0] != '0' && bookmark[0] != 'f';
    return csvFile;
}

int main(int argc_,char**argv_)
{
    if (argc_ < 2)
    {
        std::cerr << "usage: " << argv_[0] << " config-file"<< std::endl;
        return -1;
    }
    try
    {
        bridge_subscriptions subscriptions;

        std::ifstream configFile(argv_[1]);
        while(!configFile.eof() && configFile.good())
        {
            bridge_subscription subscription;
            configFile >> subscription;
            if(subscription.good())
            {
                subscriptions.push_back(subscription);
            }
        }

        bridge.run(subscriptions);
        while (1)
        {
            ::sleep(3600);
        }
    }
    catch (const AMPS::AMPSException& x_)
    {
        std::cerr << "an error occured: " << x_.toString() << std::endl;
    }
    return 0;
}
